import pandas as pd
import numbers
import numpy as np

from typing import Any, Callable, cast
from enum import Enum
from datetime import datetime
from pandas.api.types import is_bool_dtype, is_datetime64_any_dtype, is_numeric_dtype


class ParseType(str, Enum):
    """ParseType is used to specify the type of a column."""

    Boolean = "boolean"
    Date = "date"
    Integer = "int"
    Decimal = "float"
    String = "string"
    Array = "array"


def _convert_date_to_str(value: datetime, format_pattern: str) -> str | float:
    try:
        return datetime.strftime(value, format_pattern)
    except Exception:
        return np.nan


def _convert_int(value: str, radix: int) -> int | float:
    try:
        return int(value, radix)
    except ValueError:
        return np.nan


def _convert_bool(value: str) -> bool:
    return isinstance(value, str) and (value.lower() == "true")


def _to_int(column: pd.Series, radix: int) -> pd.DataFrame | pd.Series:
    if radix is None:
        if column.str.startswith("0x").any() or column.str.startswith("0X").any():
            radix = 16
        elif column.str.startswith("0").any():
            radix = 8
        else:
            radix = 10
    return column.apply(lambda x: _convert_int(x, radix))


def _convert_float(value: str) -> float:
    try:
        return float(value)
    except ValueError:
        return np.nan


def _to_str(column: pd.Series, format_pattern: str) -> pd.DataFrame | pd.Series:
    column_numeric: pd.Series | None = None
    if is_numeric_dtype(column):
        column_numeric = cast(pd.Series, pd.to_numeric(column))
    if column_numeric is not None and is_numeric_dtype(column_numeric):
        try:
            return column.apply(lambda x: "" if x is None else str(x))
        except Exception:  # noqa: S110
            pass

    try:
        datetime_column = pd.to_datetime(column)
    except Exception:
        datetime_column = column
    if is_datetime64_any_dtype(datetime_column):
        return datetime_column.apply(lambda x: _convert_date_to_str(x, format_pattern))
    if isinstance(column.dtype, pd.ArrowDtype) and "timestamp" in column.dtype.name:
        return column.apply(lambda x: _convert_date_to_str(x, format_pattern))

    if is_bool_dtype(column):
        return column.apply(lambda x: "" if pd.isna(x) else str(x).lower())
    return column.apply(lambda x: "" if pd.isna(x) else str(x))


def _to_datetime(column: pd.Series) -> pd.Series:
    if column.dropna().map(lambda x: isinstance(x, numbers.Number)).all():
        return pd.to_datetime(column, unit="ms")
    return pd.to_datetime(column)


def _to_array(column: pd.Series, delimiter: str) -> pd.Series | pd.DataFrame:
    def convert_value(value: Any) -> list:
        if pd.isna(value):
            return []
        if isinstance(value, list):
            return value
        if isinstance(value, str):
            return value.split(delimiter)
        return [value]

    return column.apply(convert_value)


__type_mapping: dict[ParseType, Callable] = {
    ParseType.Boolean: lambda column, **_kwargs: column.apply(
        lambda x: _convert_bool(x)
    ),
    ParseType.Date: lambda column, **_kwargs: _to_datetime(column),
    ParseType.Decimal: lambda column, **_kwargs: column.apply(
        lambda x: _convert_float(x)
    ),
    ParseType.Integer: lambda column, radix, **_kwargs: _to_int(column, radix),
    ParseType.String: lambda column, format_pattern, **_kwargs: _to_str(
        column, format_pattern
    ),
    ParseType.Array: lambda column, delimiter, **_kwargs: _to_array(column, delimiter),
}


def convert(
    table: pd.DataFrame,
    column: str,
    to: str,
    type: str,  # noqa: A002
    radix: int | None = None,
    delimiter: str | None = ",",
    formatPattern: str = "%Y-%m-%d",  # noqa: N803
    **_kwargs: Any,
) -> pd.DataFrame:
    """Convert verb implementation."""
    parse_type = ParseType(type)
    table[to] = __type_mapping[parse_type](
        column=table[column],
        radix=radix,
        format_pattern=formatPattern,
        delimiter=delimiter,
    )
    return table



import df_ops
import pandas as pd
from copy import deepcopy

import os
from utils.save import parquet_table_load, parquet_table_save


layout_graph_config = {"strategy": {"type": "zero"}}


def create_final_nodes(
    base_entity_graph_output: pd.DataFrame,
    query_output_dir: str,
    save: bool = True,
    try_load: bool = True,
):

    fn_name = "create_final_nodes"
    if try_load and os.path.exists(
        os.path.join(query_output_dir, f"{fn_name}.parquet")
    ):
        return parquet_table_load(query_output_dir, fn_name)
    laid_out_entity_graph = df_ops.layout_graph(
        deepcopy(base_entity_graph_output),
        **{
            "embeddings_column": "embeddings",
            "graph_column": "clustered_graph",
            "to": "node_positions",
            "graph_to": "positioned_graph",
            **layout_graph_config,
        },
    )

    nodes_without_positions = df_ops.unpack_graph(
        deepcopy(laid_out_entity_graph),
        **{"column": "positioned_graph", "type": "nodes"},
    )

    nodes_without_positions = df_ops.drop(
        nodes_without_positions,
        **{"columns": ["x", "y"]},
    )

    compute_top_level_node_positions = df_ops.unpack_graph(
        deepcopy(laid_out_entity_graph),
        **{"column": "positioned_graph", "type": "nodes"},
    )

    compute_top_level_node_positions = df_ops.filter_verb(
        compute_top_level_node_positions,
        column="level",
        value=0,
        strategy="value",
        operator="equals",
    )

    compute_top_level_node_positions = df_ops.select(
        compute_top_level_node_positions,
        **{"columns": ["id", "x", "y"]},
    )

    compute_top_level_node_positions = df_ops.rename(
        compute_top_level_node_positions,
        columns={
            "id": "top_level_node_id",
        },
    )

    compute_top_level_node_positions = df_ops.convert(
        compute_top_level_node_positions,
        **{
            "column": "top_level_node_id",
            "to": "top_level_node_id",
            "type": "string",
        },
    )

    dataset = df_ops.join(
        nodes_without_positions,
        compute_top_level_node_positions,
        on=["id", "top_level_node_id"],
    )

    dataset = df_ops.rename(
        dataset, **{"columns": {"label": "title", "cluster": "community"}}
    )

    if save:
        parquet_table_save(query_output_dir, fn_name, dataset)

    return dataset
